/*
 * HTTP 1.1 file downloader
 * (C) 2004 Sam Steele
 *
 * Email: sam.steele@gmail.com
 * Web:   http://www.c99.org/
 *
 * This source code is covered by the new BSD license.  For more info, please
 * read the LICENSE file.
 */

#include <kos.h>
#include <lwip/lwip.h>
#include <lwip/sockets.h>
#define SOCKET int

#include <string.h>
#include <math.h>
#include "http.h"

#define USER_AGENT "httptest/0.2 (KOS 1.3)"

void (*status_callback)(char *)=NULL;

void set_status_callback(http_status_callback *fn) {
  status_callback=fn;
}

int net_readline(SOCKET sock, char *buf, int bufsize) {
  int r, rt;
  char c;

  buf[0]='\0';

  rt = 0;
  do {
    r = recv(sock, &c, 1,0);
    if (r == 0)
      return -1;
    if (rt < bufsize)
      buf[rt++] = c;
  } while (c != '\n');

  buf[rt-1] = 0;
  if (buf[rt-2] == '\r')
    buf[rt-2] = 0;
  return 0;
}

int net_writeline(SOCKET sock, char *buf) {
  return send(sock,buf,strlen(buf),0);
}

int hex_to_int(char c) {
  if(c>='0' && c <='9') {
    return c-'0';
  }
  if(c>='a' && c <='f') {
    return 10+(c-'a');
  }
  if(c>='A' && c <='F') {
    return 10+(c-'A');
  }
  return -1;
}

char *strtolower(char *str) {
  int x;
  static char buf[256];
  if(str==NULL) return NULL;
  strcpy(buf,str);
  for(x=0;x<strlen(buf);x++) {
    if(buf[x]>='A' && buf[x]<='Z') {
      buf[x]+=('a'-'A');
    }
  }
  return buf;
}


void http_get_file(char *fn, const char *host, int port, const char *filename, char *content_type, int *content_length) {
  SOCKET s;
  struct sockaddr_in sinRemote;
  char *tmp;
  char name[256];
  char value[256];
  int x,y=0,z=0,fd;
  int mode=0,len=0;
  char msg[100];
  uint8 ip[4];
  uint32 host_ip;

#define READIP(dst, src) \
        dst = ((src[0]) << 24) | \
                ((src[1]) << 16) | \
                ((src[2]) << 8) | \
                ((src[3]) << 0)

  if(status_callback!=NULL) {
    sprintf(msg,"Connecting to %s...",host);
    status_callback(msg);
  }

  sinRemote.sin_family = AF_INET;
  sinRemote.sin_addr.s_addr = htonl(lwip_netcfg.dns[0]);
  sinRemote.sin_port = htons(53);

  status_callback("Resolving host...");
  lwip_gethostbyname(&sinRemote,host,ip);
  READIP(host_ip,ip);
  sprintf(msg,"%i.%i.%i.%i",ip[0],ip[1],ip[2],ip[3]);
  status_callback(msg);

  s = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
 
  sinRemote.sin_family = AF_INET;
  sinRemote.sin_addr.s_addr = htonl(host_ip);
  sinRemote.sin_port = htons(port);
  
  connect(s, (struct sockaddr*)&sinRemote, sizeof(struct sockaddr_in));

  if(status_callback!=NULL) {
    sprintf(msg,"Requesting %s...",filename);
    status_callback(msg);
  }

  tmp=malloc(256);
  sprintf(tmp,"GET %s HTTP/1.1\r\n\
Host: %s:%i\r\n\
User-Agent: %s\r\n\
Connection: keep-alive\r\n\
\r\n",filename,host,port,USER_AGENT);
  net_writeline(s,tmp);

  net_readline(s,tmp,255);
  printf("%s\n",tmp);
  strtok(tmp," "); //HTTP/1.1

  *content_length=0;
  strcpy(content_type,"unknown");

if(!strcmp("200",strtok(NULL," "))) {

  do {
    while(net_readline(s,tmp,255)==-1);
    if(tmp[0]!='\0') {
      printf("Header: %s (%i)\n",tmp,strlen(tmp));
      strcpy(name,strtolower(strtok(tmp,":")));
      strcpy(value,strtolower(strtok(NULL," ")));
      if(value[0]==' ') {
        for(x=0;x<strlen(value)-1;x++) {
          value[x]=value[x+1];
        }
      }
      if(name!=NULL) {
        if(!strcmp("transfer-encoding",name) && !strcmp("chunked",value)) {
          mode=1;
        }
        if(!strcmp("content-length",name)) {
          len=atoi(value);
          *content_length=len;
        }
        if(!strcmp("content-type",name)) {
          strcpy(content_type,value);
        }
      }
    }
  } while(tmp[0]!='\0');

  z=0;
  free(tmp);

  if(status_callback!=NULL) {
    sprintf(msg,"Downloading %s...",filename);
    status_callback(msg);
  }

  fd=fs_open(fn,O_WRONLY);

  if(mode==1) {
    *content_length=http_recieve_chunked(s,fd);
  } else if(len>0) {
    tmp=malloc(2048);
    while(y<len) {
      x=recv(s,tmp,2048,0);
      fs_write(fd,tmp,x);
      y+=x;
    }
    free(tmp);
  }
} else {
  if(status_callback!=NULL) {
    sprintf(msg,"Unable to download file");
    status_callback(msg);
  }
}
  shutdown(s,1);
  close(s);
  fs_close(fd);

  if(status_callback!=NULL) {
    sprintf(msg,"Done!");
    status_callback(msg);
  }
}

int http_recieve_chunked(SOCKET s, int fd) {
  int x,z,y,chunksize;
  char *tmp;

  y=0;
  do {
    tmp=malloc(100);

    do { //Sometimes Apache sends us a blank line, so skip over it
      net_readline(s,tmp,255); //chunk size
    } while(strlen(tmp)<1);

    for(z=0;z<strlen(tmp);z++) { //eliminate whitespace
      if(hex_to_int(tmp[z])==-1) {
        tmp[z]='\0';
        break;
      }
    }
    chunksize=0;
    if(strlen(tmp)>0) {
      for(z=0;z<strlen(tmp);z++) {
        x=pow(16,(strlen(tmp)-z-1));
        chunksize+=hex_to_int(tmp[z])*x;
      }
    }
    printf("Chunksize: %i bytes (%s)\n",chunksize,tmp);
    free(tmp);

    tmp=malloc(chunksize+1);
    if(chunksize>0) {
      x=recv(s,tmp,chunksize,0);
      fs_write(fd,tmp,x);
    }
    free(tmp);
    y+=chunksize;
  } while(chunksize>0);
  return y;
}
